#include <windows.h>
#include <LMAccess.h>

//////////////////////////////////////////////////////////////////////////////
// 
// IsAdmin  
// 
// check the token of the calling thread/process to see 
// if the caller belongs to the Administrators group.
//////////////////////////////////////////////////////////////////////////////

BOOL IsAdmin(HANDLE hToken/*=NULL*/) 
{
  BOOL   bReturn   = FALSE;
  PACL   pACL      = NULL;
  PSID   psidAdmin = NULL;
  PSECURITY_DESCRIPTOR psdAdmin = NULL;
  HANDLE hTok=NULL;
  if (hToken==NULL)
  {
    if((OpenThreadToken(GetCurrentThread(),TOKEN_DUPLICATE,FALSE,&hTok)==FALSE)
     &&(OpenProcessToken(GetCurrentProcess(),TOKEN_DUPLICATE,&hTok)==FALSE))
      return FALSE;
    hToken=hTok;
  }
  if (!DuplicateToken(hToken,SecurityImpersonation,&hToken))
    return FALSE;
  if (hTok)
    CloseHandle(hTok);
  __try 
  {
    // Initialize Admin SID and SD
    SID_IDENTIFIER_AUTHORITY SystemSidAuthority = SECURITY_NT_AUTHORITY;
    if (!AllocateAndInitializeSid(&SystemSidAuthority,2,SECURITY_BUILTIN_DOMAIN_RID,
      DOMAIN_ALIAS_RID_ADMINS,0,0,0,0,0,0,&psidAdmin))
      __leave;
    psdAdmin = LocalAlloc(LPTR,SECURITY_DESCRIPTOR_MIN_LENGTH);
    if (psdAdmin == NULL)
      __leave;
    if (!InitializeSecurityDescriptor(psdAdmin,SECURITY_DESCRIPTOR_REVISION))
      __leave;
    // Compute size needed for the ACL.
    DWORD dwACLSize=sizeof(ACL)+sizeof(ACCESS_ALLOWED_ACE)+GetLengthSid(psidAdmin)-sizeof(DWORD);
    // Allocate memory for ACL.
    pACL = (PACL)LocalAlloc(LPTR,dwACLSize);
    if (pACL == NULL)
      __leave;
    // Initialize the new ACL.
    if (!InitializeAcl(pACL,dwACLSize,ACL_REVISION2))
      __leave;
    // Add the access-allowed ACE to the DACL.
    if (!AddAccessAllowedAce(pACL,ACL_REVISION2,ACCESS_READ | ACCESS_WRITE,psidAdmin))
      __leave;
    // Set our DACL to the SD.
    if (!SetSecurityDescriptorDacl(psdAdmin,TRUE,pACL,FALSE))
      __leave;
    // AccessCheck is sensitive about what is in the SD; set the group and owner.
    SetSecurityDescriptorGroup(psdAdmin,psidAdmin,FALSE);
    SetSecurityDescriptorOwner(psdAdmin,psidAdmin,FALSE);
    if (!IsValidSecurityDescriptor(psdAdmin))
      __leave;
    // Initialize GenericMapping structure even though we won't be using generic rights.
    GENERIC_MAPPING GenericMapping;
    GenericMapping.GenericRead    = ACCESS_READ;
    GenericMapping.GenericWrite   = ACCESS_WRITE;
    GenericMapping.GenericExecute = 0;
    GenericMapping.GenericAll     = ACCESS_READ | ACCESS_WRITE;
    PRIVILEGE_SET ps;
    DWORD  dwStructureSize = sizeof(PRIVILEGE_SET);
    DWORD dwStatus;
    if (!AccessCheck(psdAdmin,hToken,ACCESS_READ,&GenericMapping,&ps, 
      &dwStructureSize,&dwStatus,&bReturn)) 
      __leave;
  } __finally 
  {
    // Cleanup 
    if (pACL) 
      LocalFree(pACL);
    if (psdAdmin) 
      LocalFree(psdAdmin);  
    if (psidAdmin) 
      FreeSid(psidAdmin);
    CloseHandle(hToken);
  }
  return bReturn;
}
