#define _WIN32_WINNT 0x0500
#define WINVER       0x0500

#include <windows.h>
#include <stdio.h>
#include <TCHAR.h>
#include <USERENV.H>
#include <Tlhelp32.h>
#include <WtsApi32.h>
#include "CmdLine.h"
#include "RunAsLib.h"

#pragma comment(lib,"advapi32")
#pragma comment(lib,"Mpr")
#pragma comment(lib,"Shell32")
#pragma comment(lib,"Userenv")
#pragma comment(lib,"Wtsapi32")
//////////////////////////////////////////////////////////////////////////////
//
//  Registry Helper
//
//////////////////////////////////////////////////////////////////////////////

BOOL GetRegStr(HKEY HK,LPCTSTR SubKey,LPCTSTR Val,LPTSTR Str,DWORD ccMax)
{
  HKEY Key;
  if (!RegOpenKeyEx(HK,SubKey,0,KEY_READ,&Key)==ERROR_SUCCESS)
    return FALSE;
  ccMax=ccMax*sizeof(TCHAR);
  DWORD dwType=0;
  BOOL bRet=(RegQueryValueEx(Key,Val,0,&dwType,(BYTE*)Str,&ccMax)==0)
          &&((dwType==REG_SZ)||(dwType==REG_EXPAND_SZ));
  RegCloseKey(Key);
  return bRet;
}

//////////////////////////////////////////////////////////////////////////////
// 
// Privilege stuff
// 
//////////////////////////////////////////////////////////////////////////////

BOOL EnablePrivilege(HANDLE hToken,LPCTSTR name)
{
  TOKEN_PRIVILEGES priv = {1,{0,0,SE_PRIVILEGE_ENABLED}};
  LookupPrivilegeValue(0,name,&priv.Privileges[0].Luid);
  AdjustTokenPrivileges(hToken,FALSE,&priv,sizeof priv,0,0);
  return  GetLastError() == ERROR_SUCCESS;
}

BOOL EnablePrivilege(LPCTSTR name)
{
  HANDLE hToken=0;
  BOOL bRet=TRUE;
  if(!OpenThreadToken(GetCurrentThread(),TOKEN_ADJUST_PRIVILEGES,0,&hToken))
    bRet&=(GetLastError()==ERROR_NO_TOKEN);
  else
  {
    bRet&=EnablePrivilege(hToken,name);
    CloseHandle(hToken);
  }
  if(!OpenProcessToken(GetCurrentProcess(),TOKEN_ADJUST_PRIVILEGES,&hToken))
    return FALSE;
  bRet&=EnablePrivilege(hToken,name);
  CloseHandle(hToken);
  return bRet;
}

//////////////////////////////////////////////////////////////////////////////
// 
// IsLocalSystem, returns true if you are running as local system
// 
//////////////////////////////////////////////////////////////////////////////

bool IsLocalSystem()
{
	HANDLE htok = 0;
	if (!OpenProcessToken(GetCurrentProcess(),TOKEN_QUERY,&htok ))
    return false;
  bool bIsLocalSystem=false;
  BYTE userSid[256];
  DWORD cb = sizeof(userSid);
  if (GetTokenInformation(htok,TokenUser,userSid,cb,&cb ))
  {
    TOKEN_USER* ptu = (TOKEN_USER*)userSid;
    SID_IDENTIFIER_AUTHORITY ntauth = SECURITY_NT_AUTHORITY;
    void* pLocalSystemSid = 0;
    if (AllocateAndInitializeSid(&ntauth,1,SECURITY_LOCAL_SYSTEM_RID,0,0,0,0,0,
                                  0,0,&pLocalSystemSid))
    {
      bIsLocalSystem=EqualSid(pLocalSystemSid,ptu->User.Sid)!=0;
      FreeSid(pLocalSystemSid);
    }
  }
  CloseHandle(htok);
	return bIsLocalSystem;
}

//////////////////////////////////////////////////////////////////////////////
//
//  NetworkPathToUNCPath
//
//  This will convert a path on a mapped network share to a UNC path
//  e.g. if you mapped \\Server\C$ to your drive Z:\ this function will
//  convert Z:\Sound\MOD4WIN\TheBest to \\Server\C$\Sound\MOD4WIN\TheBest
//  
//  This conversion is required since you loose network connections when
//  logging on as different user.
//  
//  NOTE1: The newly logged on user needs to have access permission to the
//         network share on the server
//  NOTE2: This function will fail if path is a "remembered" connection
//////////////////////////////////////////////////////////////////////////////

BOOL NetworkPathToUNCPath(LPTSTR path)
{
  if (!PathIsNetworkPath(path))
    return FALSE;
  DWORD cb=4096;
  TCHAR UNCPath[4096]={0};
  UNIVERSAL_NAME_INFO uni;
  uni.lpUniversalName=&UNCPath[0];
  UNIVERSAL_NAME_INFO* puni=(UNIVERSAL_NAME_INFO*)&UNCPath;
  DWORD dwErr=WNetGetUniversalName(path,UNIVERSAL_NAME_INFO_LEVEL,puni,&cb);
  if (dwErr!=NO_ERROR)
    return FALSE;
  _tcscpy(path,puni->lpUniversalName);
  return TRUE;
}

//////////////////////////////////////////////////////////////////////////////
//
//  RunAs:
//    Start a process with other user cedentials
//////////////////////////////////////////////////////////////////////////////

BOOL RunAs(LPCTSTR cmdline)
{
  TCHAR cmd[4096]={0};
  LPTSTR p=PathGetArgs(cmdline);
  _tcscpy(cmd,cmdline);
  PathRemoveArgs(cmd);
  PathUnquoteSpaces(cmd);
  NetworkPathToUNCPath(cmd);
  return ((DWORD)ShellExecute(0,_T("runas"),cmd,p,0,SW_SHOWNORMAL)>32);
}

//////////////////////////////////////////////////////////////////////////////
//
//  GetProcessID
//
//  Get the Process ID for a file name. Filename is stripped to the bones e.g.
//  "Explorer.exe" instead of "C:\Windows\Explorer.exe"
//  If there are multiple "Explorer.exe"s running in your Logon session, the
//  function will return the ID of the first Process it finds.
//
//  The purpose of this function ist primarily to get the Process ID of the
//  Shell process.
//////////////////////////////////////////////////////////////////////////////

DWORD GetProcessID(LPCTSTR ProcName,DWORD SessID=-1)
{
  if (SessID!=(DWORD)-1)
  {
    //Terminal Services:
    DWORD nProcesses=0;
    WTS_PROCESS_INFO* pwtspi=0;
    WTSEnumerateProcesses(WTS_CURRENT_SERVER_HANDLE,0,1,&pwtspi,&nProcesses);
    for (DWORD Process=0;Process<nProcesses;Process++) 
    {
      if (pwtspi[Process].SessionId!=SessID)
        continue;
      TCHAR PName[MAX_PATH]={0};
      _tcscpy(PName,pwtspi[Process].pProcessName);
      PathStripPath(PName);
      if (_tcsicmp(ProcName,PName)==0)
        return WTSFreeMemory(pwtspi), pwtspi[Process].ProcessId;
    }
    WTSFreeMemory(pwtspi);
  }
  //ToolHelp
  HANDLE hSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS,0);
  if (hSnap==INVALID_HANDLE_VALUE)
    return 0; 
  DWORD dwRet=0;
  PROCESSENTRY32 pe={0};
  pe.dwSize = sizeof(PROCESSENTRY32);
  bool bFirst=true;
  while((bFirst?Process32First(hSnap,&pe):Process32Next(hSnap,&pe)))
  {
    bFirst=false;
    PathStripPath(pe.szExeFile);
    if(_tcsicmp(ProcName,pe.szExeFile)!=0) 
      continue;
    if ((SessID!=(DWORD)-1))
    {
      ULONG s=-2;
      if ((!ProcessIdToSessionId(pe.th32ProcessID,&s))||(s!=SessID))
        continue;
    }
    dwRet=pe.th32ProcessID;
    break;
  }
  CloseHandle(hSnap);
  return dwRet;
}

//////////////////////////////////////////////////////////////////////////////
//
//  GetSessionUserToken
//
//  This function tries to use WTSQueryUserToken to get the token of the 
//  currently logged on user. If the WTSQueryUserToken method fails, we'll
//  try to steal the user token of the logon sessions shell process.
//////////////////////////////////////////////////////////////////////////////

HANDLE GetSessionUserToken(DWORD SessID)
{
  //WTSQueryUserToken is present in WinXP++, load it dynamically
  typedef BOOL (WINAPI* wtsqut)(ULONG,PHANDLE);
  static wtsqut wtsqueryusertoken=NULL;
  if (!wtsqueryusertoken)
  {
    HINSTANCE wtsapi32=LoadLibrary(_T("wtsapi32.dll"));
    if (wtsapi32)
      wtsqueryusertoken=(wtsqut)GetProcAddress(wtsapi32,"WTSQueryUserToken");
  }
  HANDLE hToken = NULL;
  //SE_TCB_NAME is only present in a local System User Token
  //WTSQueryUserToken requires SE_TCB_NAME!
  if ((!wtsqueryusertoken)
    ||(!EnablePrivilege(SE_TCB_NAME))
    ||(!wtsqueryusertoken(SessID,&hToken))
    ||(hToken==NULL))
  {
    //No WTSQueryUserToken: we're in Win2k
    //Get the Shells Name
    TCHAR Shell[MAX_PATH];
    if (!GetRegStr(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\Microsoft\\Windows NT\\")
                   _T("CurrentVersion\\Winlogon"),_T("Shell"),Shell,MAX_PATH))
      return 0;
    PathRemoveArgs(Shell);
    PathStripPath(Shell);
    //Now get the Shells Process ID
    DWORD ShellID=GetProcessID(Shell,SessID);
    if (!ShellID)
      return 0;
    //We got the Shells Process ID, now get the user token
    EnablePrivilege(SE_DEBUG_NAME);
    HANDLE hProc = OpenProcess(PROCESS_ALL_ACCESS,TRUE,ShellID);
    if (!hProc)
      return 0;
    // Open impersonation token for Shell process
    OpenProcessToken(hProc,TOKEN_IMPERSONATE|TOKEN_QUERY|TOKEN_DUPLICATE
                          |TOKEN_ASSIGN_PRIMARY,&hToken);
    CloseHandle(hProc);
    if(!hToken)
      return 0;
  }
  HANDLE hTokenDup=NULL;
  DuplicateTokenEx(hToken,MAXIMUM_ALLOWED,NULL,SecurityIdentification,
                          TokenPrimary,&hTokenDup); 
  CloseHandle(hToken);
  return hTokenDup;
}

//////////////////////////////////////////////////////////////////////////////
// 
//  RunAsSessionUserService
// 
//  Fake being a service.
//  Then start a command line as the user of a logon session
//////////////////////////////////////////////////////////////////////////////

const LPTSTR SName=_T("RunAsSessionUser");

static SERVICE_STATUS_HANDLE g_hSS=0;
static SERVICE_STATUS g_ss= {0};

VOID WINAPI RunSvcCtrlHndlr(DWORD dwControl)
{
  //Fake service control handler
  if(dwControl==SERVICE_CONTROL_STOP)
  {
    g_ss.dwCurrentState   = SERVICE_STOPPED; 
    g_ss.dwCheckPoint     = 0; 
    g_ss.dwWaitHint       = 0; 
    g_ss.dwWin32ExitCode  = 0; 
  } 
  if (g_hSS!=(SERVICE_STATUS_HANDLE)0) 
    SetServiceStatus(g_hSS,&g_ss);
}

VOID WINAPI RunServiceMain(DWORD argc,LPTSTR *argv)
{
  //Fake service main
  argc;//unused
  argv;//unused
  g_ss.dwServiceType      = SERVICE_WIN32_OWN_PROCESS; 
  g_ss.dwControlsAccepted = SERVICE_ACCEPT_STOP; 
  g_ss.dwCurrentState     = SERVICE_START_PENDING; 
  g_hSS                   = RegisterServiceCtrlHandler(SName,RunSvcCtrlHndlr); 
  if (g_hSS==(SERVICE_STATUS_HANDLE)0) 
    return; 
  g_ss.dwCurrentState     = SERVICE_RUNNING; 
  g_ss.dwCheckPoint       = 0;
  g_ss.dwWaitHint         = 0;
  SetServiceStatus(g_hSS,&g_ss);
  g_ss.dwCurrentState     = SERVICE_STOPPED; 
  g_ss.dwCheckPoint       = 0;
  g_ss.dwWaitHint         = 0;
  g_ss.dwWin32ExitCode    = 0; 
  SetServiceStatus(g_hSS,&g_ss);
}

bool ExitProcessIfService()
{
  //This function checks if the command line of the current process is from
  //RunAsSessionUserService. It it is, ExitProcessIfService will start and 
  //stop the fake service and then call RunAsSessionUser.
  if (!IsLocalSystem())
    return false;
  DWORD Session=-1;
  TCHAR RunCmd[4096]={0};
  {
    CCmdLine cmd;
    if ((cmd.argc()<3)||(_tcscmp(cmd.argv(1),_T("/RunInSession"))!=0))
      return false;
    Session=_ttoi(cmd.argv(2));
    for (int i=3;i<cmd.argc();i++)
    {
      if (i>3)
        _tcscat(RunCmd,_T(" "));
      TCHAR Tmp[MAX_PATH];
      _tcscpy(Tmp,cmd.argv(i));
      PathQuoteSpaces(Tmp);
      _tcscat(RunCmd,Tmp);
    }
  }
  //StartServiceCtrlDispatcher will fake a service
  SERVICE_TABLE_ENTRY DispatchTable[]={{SName,RunServiceMain},{0,0}};
  //StartServiceCtrlDispatcher is a blocking call
  StartServiceCtrlDispatcher(DispatchTable);
  //All done, delete the service
  SC_HANDLE hdlSCM = OpenSCManager(0,0,SC_MANAGER_CONNECT);
  if (hdlSCM) 
  {
    SC_HANDLE hdlServ=OpenService(hdlSCM,SName,DELETE);
    if (hdlServ)
    {
      DeleteService(hdlServ);
      CloseServiceHandle(hdlServ);
    }
  }
  //Now run the command line in the specified session and call ExitProcess
  ::ExitProcess(RunAsSessionUser(Session,RunCmd));
  return true;
}

BOOL RunAsSessionUserService(ULONG SessID,LPCTSTR CommandLine)
{
  //Create a service to run cmdline as logged on user!
  SC_HANDLE hdlSCM=OpenSCManager(0,0,SC_MANAGER_CREATE_SERVICE);
  if (hdlSCM==0) 
    return FALSE;
  //If there's us already, delete us:
  SC_HANDLE hdlServ=OpenService(hdlSCM,SName,SERVICE_STOP|DELETE);
  if (hdlServ)
  {
    DeleteService(hdlServ);
    CloseServiceHandle(hdlServ);
    hdlServ=0;
  }
  TCHAR ExeName[4096];
  TCHAR ModName[MAX_PATH];
  GetModuleFileName(NULL,ModName,MAX_PATH);
  NetworkPathToUNCPath(ModName);
  PathQuoteSpaces(ModName);
  _stprintf(ExeName,_T("%s /RunInSession %d %s"),ModName,SessID,CommandLine);
  hdlServ = CreateService(hdlSCM,SName,SName,STANDARD_RIGHTS_REQUIRED,
                          SERVICE_WIN32_OWN_PROCESS,SERVICE_DEMAND_START,
                          SERVICE_ERROR_NORMAL,ExeName,0,0,0,0,0);
  if (!hdlServ) 
    return CloseServiceHandle(hdlSCM),FALSE;
  CloseServiceHandle(hdlServ);
  hdlServ = OpenService(hdlSCM,SName,SERVICE_START);
  BOOL bRet=StartService(hdlServ,0,0);
  if (!bRet)
  {
    CloseServiceHandle(hdlServ);
    hdlServ=OpenService(hdlSCM, SName, SERVICE_STOP|DELETE);
    if (hdlServ)
      DeleteService(hdlServ);
  }
  CloseServiceHandle(hdlServ);
  CloseServiceHandle(hdlSCM);
  return bRet;
}

//////////////////////////////////////////////////////////////////////////////
// 
// The Trick: 
//  The statement below will call ExitProcessIfService() before WinMain.
//  if we restarted our process as service, WinMain will never get called
//  since ExitProcessIfService will call ::ExitProcess()
//////////////////////////////////////////////////////////////////////////////

static bool g_ExitProcessIfService=ExitProcessIfService();

//////////////////////////////////////////////////////////////////////////////
// 
//  RunAsSessionUser 
// 
//  Start CommandLine as shell user of a logon session.
//  If the current process is not running as local system, RunAsSessionUser
//  will create a service that restarts the current process as a service.
//  the service will then call RunAsSessionUser again with local system
//  credentials and exit.
//  If the current process is running as local system, RunAsSessionUser will
//  get the user token of the shell of a logon session and create a process
//  with that user token.
//////////////////////////////////////////////////////////////////////////////

DWORD RunAsSessionUser(ULONG SessID, LPCTSTR CommandLine) 
{
  //WTSQueryUserToken and CreateProcessAsUser will return "access denied" on
  //a non System account
  if (!IsLocalSystem())
    //Create a service to run cmdline!
    return RunAsSessionUserService(SessID,CommandLine);
  HANDLE hUser=GetSessionUserToken(SessID);
  if (!hUser)
    return 0;
  PROCESS_INFORMATION pi={0};
  TCHAR UserName[MAX_PATH]={0};
  PROFILEINFO ProfInf = {sizeof(ProfInf),0,UserName};
  if(GetTokenUserName(hUser,UserName) && LoadUserProfile(hUser,&ProfInf))
  {
    void* Env=0;
    if (CreateEnvironmentBlock(&Env,hUser,FALSE))
    {
      STARTUPINFO si={0};
      si.cb	= sizeof(si);
      //Do not inherit Desktop from calling process, use Tokens Desktop
      si.lpDesktop = _T("");
      //CreateProcessAsUser will only work from an NT System Account since the
      //Privilege SE_ASSIGNPRIMARYTOKEN_NAME is not present elsewhere
      EnablePrivilege(SE_ASSIGNPRIMARYTOKEN_NAME);
      EnablePrivilege(SE_INCREASE_QUOTA_NAME);
      if (CreateProcessAsUser(hUser,NULL,(LPTSTR)CommandLine,NULL,NULL,FALSE,
        CREATE_UNICODE_ENVIRONMENT|DETACHED_PROCESS,Env,NULL,&si,&pi))
      {
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
      }
      DestroyEnvironmentBlock(Env);
    }
    UnloadUserProfile(hUser,ProfInf.hProfile);
  }
  CloseHandle(hUser);
  return pi.dwProcessId;
}

//////////////////////////////////////////////////////////////////////////////
// 
//  RunAsCurrentSessionUser
// 
//////////////////////////////////////////////////////////////////////////////

DWORD RunAsCurrentSessionUser(LPCTSTR CommandLine) 
{
  ULONG SessID=0;
  if (!ProcessIdToSessionId(GetCurrentProcessId(),&SessID))
    return FALSE;
  return RunAsSessionUser(SessID,CommandLine);
}

//////////////////////////////////////////////////////////////////////////////
// 
//  GetTokenUserName
// 
//////////////////////////////////////////////////////////////////////////////

bool GetSIDUserName(PSID sid,LPTSTR Name)
{
  SID_NAME_USE snu;
  TCHAR uName[64],dName[64];
  DWORD uLen=64, dLen=64;
  if(!LookupAccountSid(NULL,sid,uName,&uLen,dName,&dLen,&snu))
    return FALSE;
  _tcscpy(Name, dName);
  if(_tcslen(Name))
    _tcscat(Name, TEXT("\\"));
  _tcscat(Name,uName);
  return TRUE;
}

bool GetTokenUserName(HANDLE hUser,LPTSTR Name)
{
  DWORD dwLen=0;
  if ((!GetTokenInformation(hUser, TokenUser,NULL,0,&dwLen))
    &&(GetLastError()!=ERROR_INSUFFICIENT_BUFFER))
    return false;
  TOKEN_USER* ptu=(TOKEN_USER*)malloc(dwLen);
  if(!ptu)
    return false;
  if(GetTokenInformation(hUser,TokenUser,(PVOID)ptu,dwLen,&dwLen))
    GetSIDUserName(ptu->User.Sid,Name);
  free(ptu);
  return true;
}

//////////////////////////////////////////////////////////////////////////////
// 
// GetShellUserName
// 
//////////////////////////////////////////////////////////////////////////////

bool GetShellUserName(LPTSTR Name)
{
  ULONG SessID=0;
  if (!ProcessIdToSessionId(GetCurrentProcessId(),&SessID))
    return FALSE;
  {
    TCHAR Shell[MAX_PATH];
    if (!GetRegStr(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\Microsoft\\Windows NT\\")
                   _T("CurrentVersion\\Winlogon"),_T("Shell"),Shell,MAX_PATH))
      return 0;
    PathRemoveArgs(Shell);
    PathStripPath(Shell);
    //Terminal Services:
    DWORD nProcesses=0;
    WTS_PROCESS_INFO* pwtspi=0;
    WTSEnumerateProcesses(WTS_CURRENT_SERVER_HANDLE,0,1,&pwtspi,&nProcesses);
    for (DWORD Process=0;Process<nProcesses;Process++) 
    {
      if (pwtspi[Process].SessionId!=SessID)
        continue;
      TCHAR PName[MAX_PATH]={0};
      _tcscpy(PName,pwtspi[Process].pProcessName);
      PathStripPath(PName);
      if (_tcsicmp(Shell,PName)==0)
        if(GetSIDUserName(pwtspi[Process].pUserSid,Name))
          return WTSFreeMemory(pwtspi),true;
    }
    WTSFreeMemory(pwtspi);
  }
  HANDLE hToken=GetSessionUserToken(SessID);
  if (!hToken)
    return FALSE;
  bool bRet=GetTokenUserName(hToken,Name);
  CloseHandle(hToken);
  return bRet;
}

//////////////////////////////////////////////////////////////////////////////
// 
// GetCurrentProcessUserName
// 
//////////////////////////////////////////////////////////////////////////////

bool GetCurrentProcessUserName(LPTSTR Name)
{
  EnablePrivilege(SE_DEBUG_NAME);
  HANDLE hProc = OpenProcess(PROCESS_ALL_ACCESS,TRUE,GetCurrentProcessId());
  if (!hProc)
    return 0;
  HANDLE hToken;
  bool bRet=false;
  // Open impersonation token for Shell process
  if (OpenProcessToken(hProc,TOKEN_QUERY,&hToken))
  {
    bRet=GetTokenUserName(hToken,Name);
    CloseHandle(hToken);
  }
  CloseHandle(hProc);
  return bRet;
}